@@ -165,7 +165,10 @@ class Document(object):
             text = self.text_after_cursor
 
         if not include_current_position:
-            text = text[1:]
+            if len(text) == 0:
+                return  # (Otherwise, we always get a match for the empty string.)
+            else:
+                text = text[1:]
 
         iterator = re.finditer(re.escape(sub), text)
 
@@ -208,7 +211,14 @@ class Document(object):
             pass
 
     def get_word_before_cursor(self):
-        return self.text_before_cursor[self.find_start_of_previous_word():]
+        """
+        Give the word before the cursor.
+        If we have whitespace before the cursor this returns an empty string.
+        """
+        if self.text_before_cursor[-1:].isspace():
+            return ''
+        else:
+            return self.text_before_cursor[self.find_start_of_previous_word():]
 
     def find_start_of_previous_word(self, count=1):
         """
@@ -464,3 +474,16 @@ class Document(object):
                     to = len(self.text)
 
             return from_, to
+
+    def empty_line_count_at_the_end(self):
+        """
+        Return number of empty lines at the end of the document.
+        """
+        count = 0
+        for line in self.lines[::-1]:
+            if not line or line.isspace():
+                count += 1
+            else:
+                break
+
+        return count
