@@ -18,6 +18,18 @@ def emacs_bindings(registry, cli_ref):
     search_line = cli_ref().lines['search']
     handle = create_handle_decorator(registry, line)
 
+    @handle(Keys.Escape)
+    def _(event):
+        """
+        By default, ignore escape key.
+
+        (If we don't put this here, and Esc is followed by a key which sequence
+        is not handled, we'll insert an Escape character in the input stream.
+        Something we don't want and happens to easily in emacs mode.
+        Further, people can always use ControlQ to do a quoted insert.)
+        """
+        pass
+
     @handle(Keys.ControlA, in_mode=InputMode.INSERT)
     @handle(Keys.ControlA, in_mode=InputMode.SELECTION)
     def _(event):
@@ -49,6 +61,13 @@ def emacs_bindings(registry, cli_ref):
         """
         line.auto_down()
 
+    @handle(Keys.ControlN, in_mode=InputMode.COMPLETE)
+    def _(event):
+        """
+        Control N in complete mode: go to next completion.
+        """
+        line.complete_next()
+
     @handle(Keys.ControlN, in_mode=InputMode.SELECTION)
     def _(event):
         """
@@ -77,6 +96,13 @@ def emacs_bindings(registry, cli_ref):
         """
         line.cursor_up()
 
+    @handle(Keys.ControlP, in_mode=InputMode.COMPLETE)
+    def _(event):
+        """
+        Control P in complete mode: go to previous completion.
+        """
+        line.complete_previous()
+
     @handle(Keys.ControlQ, Keys.Any, in_mode=InputMode.INSERT)
     def _(event):
         """
@@ -89,7 +115,8 @@ def emacs_bindings(registry, cli_ref):
         """
         Paste before cursor.
         """
-        line.paste_from_clipboard(before=True)
+        for i in range(event.arg):
+            line.paste_from_clipboard(before=True)
 
     @handle(Keys.ControlUnderscore, save_before=False, in_mode=InputMode.INSERT)
     def _(event):
@@ -166,6 +193,16 @@ def emacs_bindings(registry, cli_ref):
             words = line.document.text_after_cursor[:pos]
             line.insert_text(words.title(), overwrite=True)
 
+    @handle(Keys.Escape, 'd', in_mode=InputMode.INSERT)
+    def _(event):
+        """
+        Delete word forwards.
+        """
+        pos = line.document.find_next_word_ending(count=event.arg)
+        if pos:
+            deleted = line.delete(count=pos)
+            line.set_clipboard(ClipboardData(deleted))
+
     @handle(Keys.Escape, 'e', in_mode=InputMode.INSERT)
     def _(event):
         """ Move to end of sentence. """
