@@ -4,7 +4,7 @@ It holds the text, cursor position, history, etc...
 """
 from __future__ import unicode_literals
 
-from .code import Code, ValidationError
+from .code import Code, ValidationError, Completion
 from .document import Document
 from .enums import IncrementalSearchDirection
 from .history import History
@@ -183,13 +183,7 @@ class Line(object):
     def text(self, value):
         assert isinstance(value, six.text_type)
         self._working_lines[self.working_index] = value
-
-        # Remove validation errors, complete state.
-        self.validation_error = None
-        self.complete_state = None
-        self.selection_state = None
-
-        self.text_changed()
+        self._text_changed()
 
     @property
     def cursor_position(self):
@@ -213,16 +207,33 @@ class Line(object):
     @working_index.setter
     def working_index(self, value):
         self.__working_index = value
+        self._text_changed()
 
+    def _text_changed(self):
         # Remove any validation errors and complete state.
         self.validation_error = None
         self.complete_state = None
         self.selection_state = None
 
+        # Call text changed 'event'.
         self.text_changed()
 
     # End of <getters/setters>
 
+    @property
+    def document(self):
+        """
+        Return :class:`.Document` instance from the current text and cursor
+        position.
+        """
+        return Document(self.text, self.cursor_position, selection=self.selection_state)
+
+    def create_code(self):
+        """
+        Create `Code` instance from the current input.
+        """
+        return self.code_factory(self.document)
+
     def text_changed(self):
         """
         Not implemented. Override to capture when the current visible text
@@ -285,15 +296,6 @@ class Line(object):
             self.text[to:]
         ])
 
-    @property
-    def document(self):
-        """
-        Return :class:`.Document` instance from the current text and cursor
-        position.
-        """
-        # TODO: this can be cached as long self.text does not change.
-        return Document(self.text, self.cursor_position, selection=self.selection_state)
-
     def cursor_left(self, count=1):
         self.cursor_position += self.document.get_cursor_left_position(count=count)
 
@@ -387,18 +389,6 @@ class Line(object):
             self.working_index = index
             self.cursor_position = len(self.text)
 
-    def create_code(self):
-        """
-        Create `Code` instance from the current input.
-        """
-        return self.code_factory(self.document)
-
-    def list_completions(self):
-        """
-        Get and show all completions
-        """
-        # TODO
-
     def complete_common(self):
         """
         Autocomplete. This appends the common part of all the possible completions.
@@ -412,12 +402,14 @@ class Line(object):
         else:
             return False
 
-    def complete_next(self, count=1):
+    def complete_next(self, count=1, start_at_first=True):
         """
         Enter complete mode and browse through the completions.
+
+        :param start_at_first: If True, immediately insert the first completion.
         """
         if not self.complete_state:
-            self._start_complete()
+            self._start_complete(go_to_first=start_at_first)
         else:
             completions_count = len(self.complete_state.current_completions)
 
@@ -446,22 +438,66 @@ class Line(object):
 
             self._go_to_completion(index)
 
-    def _start_complete(self):
+    def cancel_completion(self):
+        """
+        Cancel completion, go back to the original text.
+        """
+        if self.complete_state:
+            self._go_to_completion(None)
+            self.complete_state = None
+
+    def _start_complete(self, go_to_first=True, completions=None):
         """
         Start completions. (Generate list of completions and initialize.)
         """
         # Generate list of all completions.
-        current_completions = list(self.create_code().get_completions())
+        if completions is None:
+            completions = list(self.create_code().get_completions())
 
-        if current_completions:
+        if completions:
             self.complete_state = CompletionState(
                 original_document=self.document,
-                current_completions=current_completions)
-            self._go_to_completion(0)
+                current_completions=completions)
+            if go_to_first:
+                self._go_to_completion(0)
+            else:
+                self._go_to_completion(None)
 
         else:
             self.complete_state = None
 
+    def start_history_lines_completion(self):
+        """
+        Start a completion based on all the other lines in the document and the
+        history.
+        """
+        found_completions = set()
+        completions = []
+
+        # For every line of the whole history, find matches with the current line.
+        current_line = self.document.current_line_before_cursor.lstrip()
+
+        for i, string in enumerate(self._working_lines):
+            for j, l in enumerate(string.split('\n')):
+                l = l.strip()
+                if l and l.startswith(current_line):
+                    # When a new line has been found.
+                    if l not in found_completions:
+                        found_completions.add(l)
+
+                        # Create completion.
+                        if i == self.working_index:
+                            display_meta = "Current, line %s" % (j+1)
+                        else:
+                            display_meta = "History %s, line %s" % (i+1, j+1)
+
+                        completions.append(Completion(
+                            l,
+                            start_position=-len(current_line),
+                            display_meta=display_meta))
+
+        self._start_complete(completions=completions[::-1])
+
     def _go_to_completion(self, index):
         """
         Select a completion from the list of current completions.
@@ -531,7 +567,10 @@ class Line(object):
         or return the input.
         """
         if self.is_multiline:
-            self.newline()
+            if self.document.is_cursor_at_the_end and self.document.empty_line_count_at_the_end() >= 2:
+                self.return_input()
+            else:
+                self.newline()
         else:
             self.return_input()
 
@@ -598,6 +637,18 @@ class Line(object):
         if move_cursor:
             self.cursor_position += len(data)
 
+        # Get completions.
+        if self.complete_after_insert_text():
+            self._start_complete(go_to_first=False)
+
+    def complete_after_insert_text(self):
+        """
+        Returns boolean. If True, this indicates that right now, we should get
+        the completions. E.g. for display in completion menu. (This is called
+        after ``insert_text``.)
+        """
+        return False
+
     def set_clipboard(self, clipboard_data):
         """
         Set data to the clipboard.
