@@ -5,6 +5,7 @@ from __future__ import unicode_literals
 
 from pygments.token import Token
 from .enums import IncrementalSearchDirection, InputMode
+from .renderer import Screen, Size, Point, Char
 
 __all__ = (
     'HorizontalCompletionMenu',
@@ -97,7 +98,7 @@ class HorizontalCompletionMenu(object):
                     cut_right = True
                     break
 
-            tokens.append((Token.HorizontalMenu.CurrentCompletion if i == index else Token.HorizontalMenu.Completion, c.display))
+            tokens.append((Token.HorizontalMenu.Completion.Current if i == index else Token.HorizontalMenu.Completion, c.display))
             tokens.append((Token.HorizontalMenu, ' '))
 
         # Extend/strip until the content width.
@@ -122,15 +123,24 @@ class PopupCompletionMenu(object):
     """
     Helper for drawing the complete menu to the screen.
     """
-    current_completion_token = Token.CompletionMenu.CurrentCompletion
+    current_completion_token = Token.CompletionMenu.Completion.Current
     completion_token = Token.CompletionMenu.Completion
 
+    current_meta_token = Token.CompletionMenu.Meta.Current
+    meta_token = Token.CompletionMenu.Meta
+
     progress_button_token = Token.CompletionMenu.ProgressButton
     progress_bar_token = Token.CompletionMenu.ProgressBar
 
     def __init__(self, max_height=5):
         self.max_height = max_height
 
+    def get_height(self, complete_state):
+        """
+        Return the height of the menu. (Number of rows it will use.)
+        """
+        return min(self.max_height, len(complete_state.current_completions))
+
     def write(self, screen, complete_cursor_position, complete_state):
         """
         Write the menu to the screen object.
@@ -138,17 +148,15 @@ class PopupCompletionMenu(object):
         completions = complete_state.current_completions
         index = complete_state.complete_index  # Can be None!
 
-        # Don't draw the menu if there is just one completion.
-        if len(completions) <= 1:
-            return
-
         # Get position of the menu.
         y, x = complete_cursor_position
         y += 1
         x = max(0, x - 1)  # XXX: Don't draw it in the right margin!!!...
 
         # Calculate width of completions menu.
-        menu_width = self.get_menu_width(complete_state)
+        menu_width = self.get_menu_width(screen, complete_state)
+        menu_meta_width = self.get_menu_meta_width(screen, complete_state)
+        show_meta = self.show_meta(complete_state)
 
         # Decide which slice of completions to show.
         if len(completions) > self.max_height and (index or 0) > self.max_height / 2:
@@ -179,23 +187,47 @@ class PopupCompletionMenu(object):
 
             tokens = ([(Token, ' ')] +
                       self.get_menu_item_tokens(c, is_current_completion, menu_width) +
+                      (self.get_menu_item_meta_tokens(c, is_current_completion, menu_meta_width)
+                          if show_meta else []) +
                       [(button_token, ' '), (Token, ' ')])
 
             screen.write_highlighted_at_pos(y+i, x, tokens, z_index=10)
 
-    def get_menu_width(self, complete_state):
+    def show_meta(self, complete_state):
         """
-        Calculate the menu width. This is passed to `get_menu_item_tokens`.
+        Return ``True`` if we need to show a column with meta information.
         """
-        return max(len(c.display) for c in complete_state.current_completions)
+        return any(c.display_meta for c in complete_state.current_completions)
 
-    def get_menu_item_tokens(self, completion, is_current_completion, menu_width):
+    def get_menu_width(self, screen, complete_state):
+        """
+        Return the width of the main column.
+        """
+        max_display = int(screen.size.columns / 2)
+        return min(max_display, max(len(c.display) for c in complete_state.current_completions)),
+
+    def get_menu_meta_width(self, screen, complete_state):
+        """
+        Return the width of the meta column.
+        """
+        max_display_meta = int(screen.size.columns / 4)
+        return min(max_display_meta, max(len(c.display_meta) for c in complete_state.current_completions))
+
+    def get_menu_item_tokens(self, completion, is_current_completion, width):
         if is_current_completion:
             token = self.current_completion_token
         else:
             token = self.completion_token
 
-        return [(token, ' %%-%is ' % menu_width % completion.display)]
+        return [(token, ' %%-%is ' % width % completion.display)]
+
+    def get_menu_item_meta_tokens(self, completion, is_current_completion, width):
+        if is_current_completion:
+            token = self.current_meta_token
+        else:
+            token = self.meta_token
+
+        return [(token, ' %%-%is ' % width % completion.display_meta or 'none')]
 
 
 class PasswordProcessor(object):
@@ -288,6 +320,9 @@ class Prompt(object):
     #: Text to show before the input
     prompt_text = '> '
 
+    #: Text for in the left margin in case of a multiline prompt.
+    prompt_text2 = '> '
+
     #: Processors for transforming the tokens received from the `Code` object.
     #: (This can be used for displaying password input as '*' or for
     #: highlighting mismatches of brackets in case of Python input.)
@@ -296,28 +331,38 @@ class Prompt(object):
     #: Class responsible for the composition of the i-search tokens.
     isearch_composer = ISearchComposer
 
-    def __init__(self, commandline_ref):
-        self._commandline_ref = commandline_ref
+    def __init__(self,cli_ref):
+        self._cli_ref = cli_ref
+        self.reset()
+
+    def reset(self):
+        #: Vertical scrolling position of the main content.
+        self.vertical_scroll = 0
 
     @property
-    def commandline(self):
-        return self._commandline_ref()
+    def cli(self):
+        """
+        The :class:`CommandLineInterface` instance.
+        """
+        return self._cli_ref()
 
     @property
     def line(self):
-        return self.commandline.line
+        """
+        The main :class:`Line` instance.
+        """
+        return self.cli.line
 
-    @property
-    def tokens_before_input(self):
+    def get_tokens_before_input(self):
         """
         Text shown before the actual input.
         List of (Token, text) tuples.
         """
-        if self.commandline.input_processor.input_mode == InputMode.INCREMENTAL_SEARCH and self.line.isearch_state:
+        if self.cli.input_processor.input_mode == InputMode.INCREMENTAL_SEARCH and self.line.isearch_state:
             return self.isearch_prompt
-        elif self.commandline.input_processor.input_mode == InputMode.VI_SEARCH:
+        elif self.cli.input_processor.input_mode == InputMode.VI_SEARCH:
             return self.vi_search_prompt
-        elif self.commandline.input_processor.arg is not None:
+        elif self.cli.input_processor.arg is not None:
             return self.arg_prompt
         else:
             return self.default_prompt
@@ -336,7 +381,7 @@ class Prompt(object):
         """
         return [
             (Token.Prompt.Arg, '(arg: '),
-            (Token.Prompt.Arg.Text, str(self.commandline.input_processor.arg)),
+            (Token.Prompt.Arg.Text, str(self.cli.input_processor.arg)),
             (Token.Prompt.Arg, ') '),
         ]
 
@@ -373,36 +418,23 @@ class Prompt(object):
         """
         return []
 
-    def get_tokens_in_left_margin(self, row, is_new_line):
-        """
-        When the renderer has to render the command line over several lines
-        because the input contains newline characters. This prefix will be
-        inserted before each line.
-
-        This is a generator of (Token, text) tuples.
-        """
-        # Take the length of the default prompt.
-        prompt_text = TokenList(self.tokens_before_input).text
-        length = len(prompt_text.rstrip())
-        spaces = len(prompt_text) - length
-
-        return [
-            (Token.Prompt.SecondLinePrefix, '.' * length),
-            (Token.Prompt.SecondLinePrefix, ' ' * spaces)
-        ]
-
-    def create_left_input_margin(self, screen, row, is_new_line):
-        if row > 1:
-            screen.write_highlighted(self.get_tokens_in_left_margin(row, is_new_line))
-
     def get_input_tokens(self):
-        tokens = self.line.create_code().get_tokens()
+        tokens = list(self.line.create_code().get_tokens())
 
         for p in self.input_processors:
             tokens = p.process_tokens(tokens)
 
         return tokens
 
+    def get_left_margin_tokens(self, line_number, margin_width):
+        return [
+            (Token.Prompt, self.prompt_text2)
+        ]
+
+    @property
+    def left_margin_width(self):
+        return len(self.prompt_text2)
+
     def get_highlighted_characters(self):
         """
         Return a dictionary that maps the index of input string characters to
@@ -435,41 +467,105 @@ class Prompt(object):
     def write_vi_search(self, screen):
         screen.write_highlighted(self.get_vi_search_prefix_tokens())
 
-        line = self.commandline.lines['search']
+        line = self.cli.lines['search']
 
         for index, c in enumerate(line.text + ' '):
             screen.write_char(c, Token.Prompt.ViSearch.Text,
                               set_cursor_position=(index == line.cursor_position))
 
     def write_before_input(self, screen):
-        screen.write_highlighted(self.tokens_before_input)
+        screen.write_highlighted(self.get_tokens_before_input())
 
     def write_input(self, screen, highlight=True):
-        # Set second line prefix
-        screen.set_left_margin(lambda row, is_new_line: self.create_left_input_margin(screen, row, is_new_line))
-        self.create_left_input_margin(screen, 1, True)
+        # Get tokens
+        # Note: we add the space character at the end, because that's where
+        #       the cursor can also be.
+        input_tokens = self.get_input_tokens() + [(Token, ' ')]
 
+        # 'Explode' tokens in characters.
+        input_tokens = [(token, c) for token, text in input_tokens for c in text]
+
+        # Apply highlighting.
         if highlight:
             highlighted_characters = self.get_highlighted_characters()
-        else:
-            highlighted_characters = {}
 
-        index = 0
-        # Note, we add the space character at the end, because that's where
-        #       the cursor could be.
-        for token, text in self.get_input_tokens() + [(Token, ' ')]:
-            for c in text:
-                # Determine Token-type for character.
-                t = highlighted_characters.get(index, token)
+            for index, token in highlighted_characters.items():
+                input_tokens[index] = (token, input_tokens[index][1])
 
-                # Insert char.
-                screen.write_char(c, t,
-                                  string_index=index,
-                                  set_cursor_position=(index == self.line.cursor_position))
-                index += 1
+        for index, (token, c) in enumerate(input_tokens):
+            # Insert char.
+            screen.write_char(c, token,
+                              string_index=index,
+                              set_cursor_position=(index == self.line.cursor_position))
 
-        # Unset left margin.
-        screen.set_left_margin(None)
+    def write_input_scrolled(self, screen, accept_or_abort=False, min_available_height=1, left_margin_width=0,
+            bottom_margin_height=0):
+        """
+        Write visible part of the input to the screen. (Scroll if the input is
+        too large.)
+
+        :return: Cursor row position after the scroll region.
+        """
+        # Write to a temp screen first. (Later, we will copy the visible region
+        # of this screen to the real screen.)
+        temp_screen = Screen(Size(columns=screen.size.columns - left_margin_width,
+                                  rows=screen.size.rows))
+        self.write_input(temp_screen, highlight=not accept_or_abort)
+
+        # Determine the maximum height.
+        max_height = screen.size.rows - 2
+
+        # Scroll.
+        if True:
+            # Scroll back if we scrolled to much and there's still space at the top.
+            if self.vertical_scroll > temp_screen.current_height - max_height:
+                self.vertical_scroll = max(0, temp_screen.current_height - max_height)
+
+            # Scroll up if cursor is before visible part.
+            if self.vertical_scroll > temp_screen.cursor_position.y:
+                self.vertical_scroll = temp_screen.cursor_position.y
+
+            # Scroll down if cursor is after visible part.
+            if self.vertical_scroll <= temp_screen.cursor_position.y - max_height:
+                self.vertical_scroll = (temp_screen.cursor_position.y + 1) - max_height
+
+            # Scroll down if we need space for the menu.
+            if self.need_to_show_completion_menu():
+                menu_size = self.completion_menu.get_height(self.line.complete_state) - 1
+                if temp_screen.cursor_position.y - self.vertical_scroll >= max_height - menu_size:
+                    self.vertical_scroll = (temp_screen.cursor_position.y + 1) - (max_height - menu_size)
+
+        # Now copy the region we need to the real screen.
+        y = 0
+        for y in range(0, min(max_height, temp_screen.current_height - self.vertical_scroll)):
+            # Write line numbers.
+            screen.write_highlighted_at_pos(y, 0, self.get_left_margin_tokens(
+                y + self.vertical_scroll, left_margin_width))
+
+            # Write line content.
+            for x in range(0, temp_screen.size.columns):
+                screen._buffer[y][x + left_margin_width] = temp_screen._buffer[y + self.vertical_scroll][x]
+
+        screen.cursor_position = Point(y=temp_screen.cursor_position.y - self.vertical_scroll,
+                                       x=temp_screen.cursor_position.x + left_margin_width)
+
+        y_after_input = y
+
+        # Show completion menu.
+        if not accept_or_abort and self.need_to_show_completion_menu():
+            y, x = temp_screen._cursor_mappings[self.line.complete_state.original_document.cursor_position]
+            self.completion_menu.write(screen, (y - self.vertical_scroll, x + left_margin_width), self.line.complete_state)
+
+        return_value = max([min_available_height, screen.current_height]) - bottom_margin_height
+
+        # Fill up with tildes.
+        if not accept_or_abort:
+            y = y_after_input + 1
+            while y < max([min_available_height, screen.current_height]) - bottom_margin_height:
+                screen.write_at_pos(y, 1, Char('~', Token.Leftmargin.Tilde))
+                y += 1
+
+        return return_value
 
     def write_after_input(self, screen):
         """
@@ -478,8 +574,7 @@ class Prompt(object):
         screen.write_highlighted(self.get_tokens_after_input())
 
     def need_to_show_completion_menu(self):
-        return (self.commandline.input_processor.input_mode == InputMode.COMPLETE and
-                self.completion_menu and self.line.complete_state)
+        return self.completion_menu and self.line.complete_state
 
     def write_menus(self, screen):
         """
@@ -491,15 +586,31 @@ class Prompt(object):
 
             self.completion_menu.write(screen, complete_cursor_position, self.line.complete_state)
 
-    def write_to_screen(self, screen, last_screen_height, accept=False, abort=False):
+    def write_to_screen(self, screen, min_available_height, accept=False, abort=False):
         """
         Render the prompt to a `Screen` instance.
+
+        :param screen: The :class:`Screen` class into which we write the output.
+        :param min_available_height: The space (amount of rows) available from
+                                     the top of the prompt, until the bottom of
+                                     the terminal. We don't have to use them,
+                                     but we can.
         """
-        if self.commandline.input_processor.input_mode == InputMode.VI_SEARCH:
+        if self.cli.input_processor.input_mode == InputMode.VI_SEARCH:
             self.write_vi_search(screen)
         else:
             self.write_before_input(screen)
-            self.write_input(screen, highlight=not (accept or abort))
+
+            # Write actual input.
+            if self.line.is_multiline:
+                self.write_input_scrolled(screen,
+                                      accept_or_abort=(accept or abort),
+                                      min_available_height=min_available_height,
+                                      left_margin_width=self.left_margin_width)
+            else:
+                self.write_input(screen, highlight=not (accept or abort))
+
+
 
         if not (accept or abort):
             self.write_after_input(screen)
